#include <unistd.h>
#include <stdlib.h>
#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>
#include <glib/gi18n.h>
#include <glib-object.h>
#include "e-data-book-enumtypes.h"
#include "e-data-book-factory.h"
#include "e-data-book.h"
#include "e-data-book-view.h"
#include "e-book-backend-sexp.h"

extern DBusGConnection *connection;

/* DBus glue */
static gboolean impl_AddressBook_Book_open(EDataBook *book, gboolean only_if_exists, DBusGMethodInvocation *context);
static gboolean impl_AddressBook_Book_remove(EDataBook *book, DBusGMethodInvocation *context);
static gboolean impl_AddressBook_Book_getContact(EDataBook *book, const char *IN_uid, DBusGMethodInvocation *context);
static gboolean impl_AddressBook_Book_getContactList(EDataBook *book, const char *query, DBusGMethodInvocation *context);
static gboolean impl_AddressBook_Book_authenticateUser(EDataBook *book, const char *IN_user, const char *IN_passwd, const char *IN_auth_method, DBusGMethodInvocation *context);
static gboolean impl_AddressBook_Book_addContact(EDataBook *book, const char *IN_vcard, DBusGMethodInvocation *context);
static gboolean impl_AddressBook_Book_modifyContact(EDataBook *book, const char *IN_vcard, DBusGMethodInvocation *context);
static gboolean impl_AddressBook_Book_removeContacts(EDataBook *book, const char **IN_uids, DBusGMethodInvocation *context);
static gboolean impl_AddressBook_Book_getStaticCapabilities(EDataBook *book, char **OUT_capabilities, GError **error);
static gboolean impl_AddressBook_Book_getSupportedFields(EDataBook *book, DBusGMethodInvocation *context);
static gboolean impl_AddressBook_Book_getRequiredFields(EDataBook *book, DBusGMethodInvocation *context);
static gboolean impl_AddressBook_Book_getSupportedAuthMethods(EDataBook *book, DBusGMethodInvocation *context);
static gboolean impl_AddressBook_Book_getBookView (EDataBook *book, const char *address, const char *listener_path, const char *search, const char **fields, const guint max_results, char **view_path, GError **error);
static gboolean impl_AddressBook_Book_getChanges(EDataBook *book, const char *IN_change_id, DBusGMethodInvocation *context);
static gboolean impl_AddressBook_Book_cancelOperation(EDataBook *book, GError **error);
static gboolean impl_AddressBook_Book_close(EDataBook *object, GError **error);
#include "e-data-book-glue.h"

static void return_status_and_list (guint32 opid, EDataBookStatus status, GList *list, gboolean free_data);

/* Convenience macro to test and set a GError/return on failure */
#define g_set_error_val_if_fail(test, returnval, error, domain, code) G_STMT_START{ \
 if G_LIKELY (test) {} else { \
   g_set_error (error, domain, code, #test); \
   g_warning(#test " failed"); \
   return (returnval); \
 } \
}G_STMT_END

enum
{
  WRITABLE,
  CONNECTION,
  AUTH_REQUIRED,
  LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = { 0 };

/* Generate the GObject boilerplate */
G_DEFINE_TYPE(EDataBook, e_data_book, G_TYPE_OBJECT)

/* Create the EDataBook error quark */
GQuark
e_data_book_error_quark (void)
{
  static GQuark quark = 0;
  if (!quark)
    quark = g_quark_from_static_string ("e_data_book_error");
  return quark;
}

/* Class init */
static void e_data_book_dispose (GObject *object);
static void
e_data_book_class_init (EDataBookClass *e_data_book_class)
{ 
  GObjectClass *object_class = G_OBJECT_CLASS (e_data_book_class);

  object_class->dispose = e_data_book_dispose;

  signals[WRITABLE] =
    g_signal_new ("writable",
                  G_OBJECT_CLASS_TYPE (e_data_book_class),
                  G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                  0,
                  NULL, NULL,
                  g_cclosure_marshal_VOID__BOOLEAN,
                  G_TYPE_NONE, 1, G_TYPE_BOOLEAN);

  signals[CONNECTION] =
    g_signal_new ("connection",
                  G_OBJECT_CLASS_TYPE (e_data_book_class),
                  G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                  0,
                  NULL, NULL,
                  g_cclosure_marshal_VOID__BOOLEAN,
                  G_TYPE_NONE, 1, G_TYPE_BOOLEAN);

  signals[AUTH_REQUIRED] =
    g_signal_new ("auth-required",
                  G_OBJECT_CLASS_TYPE (e_data_book_class),
                  G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                  0,
                  NULL, NULL,
                  g_cclosure_marshal_VOID__VOID,
                  G_TYPE_NONE, 0);

  dbus_g_object_type_install_info (G_TYPE_FROM_CLASS (e_data_book_class), &dbus_glib_e_data_book_object_info);
  
  dbus_g_error_domain_register (E_DATA_BOOK_ERROR, NULL, E_TYPE_DATA_BOOK_STATUS);
}

/* Instance init */
static void
e_data_book_init (EDataBook *ebook)
{
}

EDataBook *
e_data_book_new (EBookBackend *backend, ESource *source)
{
  EDataBook *book;
  book = g_object_new (E_TYPE_DATA_BOOK, NULL);
  book->backend = g_object_ref (backend);
  book->source = g_object_ref (source);
  return book;
}

static void
e_data_book_dispose (GObject *object)
{
  EDataBook *book = E_DATA_BOOK (object);

  g_list_foreach (book->views, g_object_unref, NULL);
  g_list_free (book->views);

  g_object_unref (book->backend);
  g_object_unref (book->source);

  if (G_OBJECT_CLASS (e_data_book_parent_class)->dispose)
    G_OBJECT_CLASS (e_data_book_parent_class)->dispose (object);	
}

ESource*
e_data_book_get_source (EDataBook *book)
{
  return book->source;
}

EBookBackend*
e_data_book_get_backend (EDataBook *book)
{
  return book->backend;
}

static gboolean
impl_AddressBook_Book_open(EDataBook *book, gboolean only_if_exists, DBusGMethodInvocation *context)
{
  e_book_backend_open (book->backend, book, GPOINTER_TO_INT (context), only_if_exists);
  return TRUE;
}

void
e_data_book_respond_open (EDataBook *book, guint opid, EDataBookStatus status)
{
#if 0
  if (status != Success) {
    dbus_g_method_return_error (GINT_TO_POINTER (opid), g_error_new (E_DATA_BOOK_ERROR, status, _("Cannot open book")));
  } else {
    dbus_g_method_return (GINT_TO_POINTER (opid));
  }
#else
  dbus_g_method_return (GINT_TO_POINTER (opid), status);
#endif
}

static gboolean
impl_AddressBook_Book_remove(EDataBook *book, DBusGMethodInvocation *context)
{
  e_book_backend_remove (book->backend, book, GPOINTER_TO_INT (context));
  return TRUE;
}

void
e_data_book_respond_remove (EDataBook *book, guint opid, EDataBookStatus status)
{
  if (status != Success) {
    dbus_g_method_return_error (GINT_TO_POINTER (opid), g_error_new (E_DATA_BOOK_ERROR, status, _("Cannot remove book")));
  } else {
    dbus_g_method_return (GINT_TO_POINTER (opid));
  }
}

static gboolean
impl_AddressBook_Book_getContact(EDataBook *book, const char *IN_uid, DBusGMethodInvocation *context)
{
  //g_set_error_val_if_fail (IN_uid != NULL, FALSE, error, E_DATA_BOOK_ERROR, ContactNotFound);
  e_book_backend_get_contact (book->backend, book, GPOINTER_TO_INT (context), IN_uid);
  return TRUE;
}

void
e_data_book_respond_get_contact (EDataBook *book, guint32 opid, EDataBookStatus status, char *vcard)
{
  if (status != Success) {
    dbus_g_method_return_error (GINT_TO_POINTER (opid), g_error_new (E_DATA_BOOK_ERROR, status, _("Cannot get contact")));
  } else {
    dbus_g_method_return (GINT_TO_POINTER (opid), vcard);
  }
}

static gboolean
impl_AddressBook_Book_getContactList(EDataBook *book, const char *query, DBusGMethodInvocation *context)
{
  e_book_backend_get_contact_list (book->backend, book, GPOINTER_TO_INT (context), query);
  return TRUE;
}

void
e_data_book_respond_get_contact_list (EDataBook *book, guint32 opid, EDataBookStatus status, GList *cards)
{
  return_status_and_list (opid, status, cards, TRUE);
}

static gboolean
impl_AddressBook_Book_authenticateUser(EDataBook *book, const char *IN_user, const char *IN_passwd, const char *IN_auth_method, DBusGMethodInvocation *context)
{
  e_book_backend_authenticate_user (e_data_book_get_backend (book), book,
                                    GPOINTER_TO_INT (context), IN_user, IN_passwd, IN_auth_method);

  return TRUE;
}

void
e_data_book_respond_authenticate_user (EDataBook *book, guint32 opid, EDataBookStatus status)
{
  if (status != Success) {
    dbus_g_method_return_error (GINT_TO_POINTER (opid), g_error_new (E_DATA_BOOK_ERROR, status, _("Cannot authenticate user")));
  } else {
    dbus_g_method_return (GINT_TO_POINTER (opid));
  }
}

static gboolean
impl_AddressBook_Book_addContact(EDataBook *book, const char *IN_vcard, DBusGMethodInvocation *context)
{
  e_book_backend_create_contact (e_data_book_get_backend (book), book,
                                 GPOINTER_TO_INT (context), IN_vcard);

  return TRUE;
}

void
e_data_book_respond_create (EDataBook *book, guint32 opid, EDataBookStatus status, EContact *contact)
{
  if (status == Success) {
    e_book_backend_notify_update (e_data_book_get_backend (book), contact);
    e_book_backend_notify_complete (e_data_book_get_backend (book));
  }

  if (status != Success) {
    dbus_g_method_return_error (GINT_TO_POINTER (opid), g_error_new (E_DATA_BOOK_ERROR, status, _("Cannot add contact")));
  } else {
    dbus_g_method_return (GINT_TO_POINTER (opid), e_contact_get_const (contact, E_CONTACT_UID));
  }
}

static gboolean
impl_AddressBook_Book_modifyContact(EDataBook *book, const char *IN_vcard, DBusGMethodInvocation *context)
{
  e_book_backend_modify_contact (e_data_book_get_backend (book), book,
                                 GPOINTER_TO_INT (context), IN_vcard);

  return TRUE;
}

void
e_data_book_respond_modify (EDataBook *book, guint32 opid, EDataBookStatus status, EContact *contact)
{
  if (status == Success) {
    e_book_backend_notify_update (e_data_book_get_backend (book), contact);
    e_book_backend_notify_complete (e_data_book_get_backend (book));
  }

  if (status != Success) {
    dbus_g_method_return_error (GINT_TO_POINTER (opid), g_error_new (E_DATA_BOOK_ERROR, status, _("Cannot modify contact")));
  } else {
    dbus_g_method_return (GINT_TO_POINTER (opid));
  }
}

static gboolean
impl_AddressBook_Book_removeContacts(EDataBook *book, const char **IN_uids, DBusGMethodInvocation *context)
{
  GList *id_list = NULL;
  int i = 0;

  while (IN_uids[i] != NULL) {
    id_list = g_list_prepend (id_list, (gpointer) IN_uids[i]);
    i++;
  }
  
  e_book_backend_remove_contacts (e_data_book_get_backend (book), book, 
                                  GPOINTER_TO_INT (context), id_list);
  g_list_free (id_list);

  return TRUE;
}

void
e_data_book_respond_remove_contacts (EDataBook *book, guint32 opid, EDataBookStatus status, GList *ids)
{
  if (ids) {
    GList *i;

    for (i = ids; i; i = i->next)
      e_book_backend_notify_remove (e_data_book_get_backend (book), i->data);
    e_book_backend_notify_complete (e_data_book_get_backend (book));
  }

  if (status != Success) {
    dbus_g_method_return_error (GINT_TO_POINTER (opid), g_error_new (E_DATA_BOOK_ERROR, status, _("Cannot remove contacts")));
  } else {
    dbus_g_method_return (GINT_TO_POINTER (opid));
  }
}

static gboolean
impl_AddressBook_Book_getStaticCapabilities(EDataBook *book, char **OUT_capabilities, GError **error)
{
  *OUT_capabilities = e_book_backend_get_static_capabilities (e_data_book_get_backend (book));
  return TRUE;
}

static gboolean
impl_AddressBook_Book_getSupportedFields(EDataBook *book, DBusGMethodInvocation *context)
{
  e_book_backend_get_supported_fields (e_data_book_get_backend (book), book, GPOINTER_TO_INT (context));
  return TRUE;
}

void
e_data_book_respond_get_supported_fields (EDataBook *book, guint32 opid, EDataBookStatus status, GList *fields)
{
  return_status_and_list (opid, status, fields, FALSE);
}

static gboolean
impl_AddressBook_Book_getRequiredFields(EDataBook *book, DBusGMethodInvocation *context)
{
  e_book_backend_get_required_fields (e_data_book_get_backend (book), book, GPOINTER_TO_INT (context));
  return TRUE;
}

void
e_data_book_respond_get_required_fields (EDataBook *book, guint32 opid, EDataBookStatus status, GList *fields)
{
  return_status_and_list (opid, status, fields, FALSE);
}

static gboolean
impl_AddressBook_Book_getSupportedAuthMethods(EDataBook *book, DBusGMethodInvocation *context)
{
  e_book_backend_get_supported_auth_methods (e_data_book_get_backend (book), book, GPOINTER_TO_INT (context));
  return TRUE;
}

void
e_data_book_respond_get_supported_auth_methods (EDataBook *book, guint32 opid, EDataBookStatus status, GList *auth_methods)
{
  return_status_and_list (opid, status, auth_methods, FALSE);
}

static char*
construct_bookview_path (void)
{
  static guint counter = 1;
  return g_strdup_printf ("/org/gnome/evolution/dataserver/addressbook/BookView/%d/%d", getpid(), counter++);
}

static void
view_destroy(gpointer data, GObject *dead)
{
  EDataBook *book = E_DATA_BOOK (data);
  e_book_backend_remove_book_view (book->backend, (EDataBookView*)dead);
  book->views = g_list_remove (book->views, dead);
}

static gboolean
impl_AddressBook_Book_getBookView (EDataBook *book, const char *address, const char *listener_path, const char *search, const char **fields, const guint max_results, char **view_path, GError **error)
{
  EBookBackend *backend = e_data_book_get_backend (book);
  EBookBackendSExp *card_sexp;
  EDataBookView *book_view;
  DBusGProxy *listener;
  char *path;

  card_sexp = e_book_backend_sexp_new (search);
  if (!card_sexp) {
    g_set_error (error, E_DATA_BOOK_ERROR, InvalidQuery, "Invalid query");
    return FALSE;
  }

  listener = dbus_g_proxy_new_for_name (connection, address, listener_path, "org.gnome.evolution.dataserver.addressbook.DataBookViewListener");
  g_assert (listener != NULL);
  book_view = e_data_book_view_new (backend, listener, search, card_sexp, max_results);
  g_object_weak_ref (G_OBJECT (book_view), view_destroy, book);
  path = construct_bookview_path ();
  dbus_g_connection_register_g_object (connection, path, G_OBJECT (book_view));
  e_book_backend_add_book_view (backend, book_view);
  *view_path = path;
  book->views = g_list_prepend(book->views, book_view);
  return TRUE;
}

static gboolean
impl_AddressBook_Book_getChanges(EDataBook *book, const char *IN_change_id, DBusGMethodInvocation *context)
{
  e_book_backend_get_changes (e_data_book_get_backend (book), book, GPOINTER_TO_INT (context), IN_change_id);
  return TRUE;
}

void
e_data_book_respond_get_changes (EDataBook *book, guint32 opid, EDataBookStatus status, GList *changes)
{
  char **list;
  int i = 0;
  /* TODO: can probably assume that on !Success list is always NULL */
  /* DBus wants a char**, so convert */
  list = g_new0 (char*, g_list_length (changes) + 1);
  while (changes != NULL) {
    EDataBookChange *change = (EDataBookChange *) changes->data;
    list[i++] = g_strdup_printf ("%d\n%s", change->change_type, change->vcard);
    g_free (change->vcard);
    g_free (change);
    changes = g_list_remove (changes, change);
  }
  if (status != Success) {
    dbus_g_method_return_error (GINT_TO_POINTER (opid), g_error_new (E_DATA_BOOK_ERROR, status, _("Cannot get changes")));
  } else {
    dbus_g_method_return (GINT_TO_POINTER (opid), list);
  }
  g_strfreev (list);
}

static gboolean
impl_AddressBook_Book_cancelOperation(EDataBook *book, GError **error)
{
  if (!e_book_backend_cancel_operation (e_data_book_get_backend (book), book)) {
    g_set_error (error, E_DATA_BOOK_ERROR, CouldNotCancel, "Failed to cancel operation");
    return FALSE;
  }

  return TRUE;
}

static gboolean
impl_AddressBook_Book_close(EDataBook *book, GError **error)
{
  g_object_unref (book);
  return TRUE;
}

void
e_data_book_report_writable (EDataBook *book, gboolean writable)
{
  g_signal_emit (book, signals[WRITABLE], 0, writable);
}

void
e_data_book_report_connection_status (EDataBook *book, gboolean connected)
{
  g_signal_emit (book, signals[CONNECTION], 0, connected);
}

void
e_data_book_report_auth_required (EDataBook *book)
{
  g_signal_emit (book, signals[AUTH_REQUIRED], 0);
}

static void
return_status_and_list (guint32 opid, EDataBookStatus status, GList *list, gboolean free_data)
{
  if (status == Success) {
    char **array;
    GList *l;
    int i = 0;
    
    array = g_new0 (char*, g_list_length (list) + 1);
    for (l = list; l != NULL; l = l->next) {
      array[i++] = l->data;
    }
    
    dbus_g_method_return (GINT_TO_POINTER (opid), array);
    
    if (free_data) {
      g_strfreev (array);
    } else {
      g_free (array);
    }
  } else {
    dbus_g_method_return_error (GINT_TO_POINTER (opid), g_error_new (E_DATA_BOOK_ERROR, status, _("Cannot complete operation")));
  }
}

/*
TODO: finalise dispose
*/
